/**
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

/**
 **  gcspy_interpreter.cpp
 **
 **  The interpreter communicating with the client
 **/

#include <stdio.h>
#include "gcspy_interpreter.hpp"

/* FIXME g++ 4.1.2 complains that ‘pause_req_cmd’ was not declared in this scope
 * if it's used in this file before it's declared!!
gcspyInterpreter::gcspyInterpreter (gcspyClient *client)
  : gcspyCommandStream(cmd_len) {
  _client = client;

  _cmds[pause_req_tag]      = pause_req_cmd;
  _cmds[pause_tag]          = pause_cmd;
  _cmds[restart_tag]        = restart_cmd;
  _cmds[play_one_tag]       = play_one_cmd;
  _cmds[shutdown_req_tag]   = shutdown_req_cmd;
  _cmds[shutdown_tag]       = shutdown_cmd;
  _cmds[stream_tag]         = stream_cmd;
  _cmds[event_tag]          = event_cmd;
  _cmds[control_tag]        = control_cmd;
  _cmds[filter_tag]         = filter_cmd;
  _cmds[event_count_tag]    = event_count_cmd;
  _cmds[summary_tag]        = summary_cmd;
  _cmds[space_info_tag]     = space_info_cmd;
  _cmds[space_tag]          = space_cmd;
}
*/

void
gcspyInterpreter::send_single_cmd (unsigned tag) {
  gcspyBufferedOutput output = gcspyBufferedOutput(_client);

  start(&output);
  put_tag(&output, tag);
  finish(&output);
}

////////// PAUSE REQ //////////

void
pause_req_cmd (gcspyBufferedInput *input, void *aux) {
  gcspyServerState *state = (gcspyServerState *) aux;
  state->set_pause_req();
}

////////// PAUSE //////////

void
pause_cmd (gcspyBufferedInput *input, void *aux) {
  gcspy_raise_error("PAUSE REQ command not implemented (not necessary)");
}

void
gcspyInterpreter::send_pause_cmd () {
  send_single_cmd(pause_tag);
}

////////// RESTART //////////

void
restart_cmd (gcspyBufferedInput *input, void *aux) {
  gcspyServerState *state = (gcspyServerState *) aux;
  state->set_restart();
}

////////// PLAY ONE //////////

void
play_one_cmd (gcspyBufferedInput *input, void *aux) {
  gcspyServerState *state = (gcspyServerState *) aux;
  state->set_play_one();
}

////////// SHUTDOWN REQ //////////

void
shutdown_req_cmd (gcspyBufferedInput *input, void *aux) {
  gcspyServerState *state = (gcspyServerState *) aux;
  state->set_shutdown_req();
}

////////// SHUTDOWN //////////

void
shutdown_cmd (gcspyBufferedInput *input, void *aux) {
  gcspy_raise_error("SHUTDOWN command not implemented (not necessary)");
}

void
gcspyInterpreter::send_shutdown_cmd () {
  send_single_cmd(shutdown_tag);
}

////////// STREAM //////////

void
stream_cmd (gcspyBufferedInput *input, void *aux) {
  gcspy_raise_error("STREAM command not implemented (not necessary)");
}

void
gcspyInterpreter::send_stream_cmd (unsigned space_id,
				   unsigned stream_id,
				   unsigned data_type,
				   char *data,
				   gcspy_length_t len) {
  gcspyBufferedOutput output = gcspyBufferedOutput(_client);

  start(&output);
  put_tag(&output, stream_tag);
  output.write_byte((char) space_id);
  output.write_byte((char) stream_id);
  switch (data_type) {
  case gcspyStream::byte_type:
    output.write_byte_array(data, len);
    break;
  case gcspyStream::short_type:
    output.write_short_array((short *) data, len);
    break;
  case gcspyStream::int_type:
    output.write_int_array((int *) data, len);
    break;
  default:
    gcspy_assert(0);
    break;
  }
  finish(&output);
}

////////// EVENT //////////

void
event_cmd (gcspyBufferedInput *input, void *aux) {
  gcspy_raise_error("EVENT command not implemented (not necessary)");
}

void
gcspyInterpreter::send_event_cmd (unsigned event,
				  int elapsed_time,
				  int compensation_time) {

  gcspyBufferedOutput output = gcspyBufferedOutput(_client);

  start(&output);
  put_tag(&output, event_tag);
  output.write_byte((char) event);
  output.write_int(elapsed_time);
  output.write_int(compensation_time);
  finish(&output);
}

////////// CONTROL //////////

void
control_cmd (gcspyBufferedInput *input, void *aux) {
  gcspy_raise_error("CONTROL command not implemented (not necessary)");
}

void
gcspyInterpreter::send_control_cmd (unsigned space_id,
				    char *data,
				    gcspy_length_t len) {
  gcspyBufferedOutput output = gcspyBufferedOutput(_client);

  start(&output);
  put_tag(&output, control_tag);
  output.write_byte((char) space_id);
  output.write_byte_array(data, len);
  finish(&output);
}

////////// FILTER //////////

void
filter_cmd (gcspyBufferedInput *input, void *aux) {
  gcspyServerState *state = (gcspyServerState *) aux;

  gcspy_bool_t *enabled = state->get_enabled_array();
  int *delays = state->get_delays_array();
  gcspy_bool_t *paused = state->get_paused_array();
  int *periods = state->get_periods_array();

  gcspy_length_t len = (gcspy_length_t) input->read_short();
  for (int i = 0; i < len; ++i) {
    enabled[i] = input->read_boolean();
    delays[i] = input->read_int();
    paused[i] = input->read_boolean();
    periods[i] = input->read_int();
  }
}

////////// EVENT COUNT //////////

void
event_count_cmd (gcspyBufferedInput *input, void *aux) {
  gcspy_raise_error("EVENT COUNT command not implemented (not necessary)");
}

void
gcspyInterpreter::send_event_count_cmd (int *event_counts,
					gcspy_length_t len) {
  gcspyBufferedOutput output = gcspyBufferedOutput(_client);

  start(&output);
  put_tag(&output, event_count_tag);
  output.write_int_array(event_counts, len);
  finish(&output);
}

////////// SUMMARY //////////

void
summary_cmd (gcspyBufferedInput *input, void *aux) {
  gcspy_raise_error("SUMMARY command not implemented (not necessary)");
}

void
gcspyInterpreter::send_summary_cmd (unsigned space_id,
				    unsigned stream_id,
				    int *summary,
				    gcspy_length_t len) {
  gcspyBufferedOutput output = gcspyBufferedOutput(_client);

  start(&output);
  put_tag(&output, summary_tag);
  output.write_byte((char) space_id);
  output.write_byte((char) stream_id);
  if (summary == NULL)
    output.write_empty_array();
  else
    output.write_int_array(summary, len);
  finish(&output);
}

////////// SPACE INFO //////////

void
space_info_cmd (gcspyBufferedInput *input, void *aux) {
  gcspy_raise_error("SPACE INFO command not implemented (not necessary)");
}

void
gcspyInterpreter::send_space_info_cmd (unsigned space_id,
				       char *space_info) {
  gcspyBufferedOutput output = gcspyBufferedOutput(_client);

  start(&output);
  put_tag(&output, space_info_tag);
  output.write_byte((char) space_id);
  output.write_string(space_info);
  finish(&output);
}

////////// SPACE //////////

void
space_cmd (gcspyBufferedInput *input, void *aux) {
  gcspy_raise_error("SPACE command not implemented (not necessary)");
}

void
gcspyInterpreter::send_space_cmd (gcspySpace *space) {
  gcspyBufferedOutput output = gcspyBufferedOutput(_client);

  start(&output);
  put_tag(&output, space_tag);
  space->serialise(&output);
  finish(&output);
}

gcspyInterpreter::gcspyInterpreter (gcspyClient *client)
  : gcspyCommandStream(cmd_len) {
  _client = client;

  _cmds[pause_req_tag]      = pause_req_cmd;
  _cmds[pause_tag]          = pause_cmd;
  _cmds[restart_tag]        = restart_cmd;
  _cmds[play_one_tag]       = play_one_cmd;
  _cmds[shutdown_req_tag]   = shutdown_req_cmd;
  _cmds[shutdown_tag]       = shutdown_cmd;
  _cmds[stream_tag]         = stream_cmd;
  _cmds[event_tag]          = event_cmd;
  _cmds[control_tag]        = control_cmd;
  _cmds[filter_tag]         = filter_cmd;
  _cmds[event_count_tag]    = event_count_cmd;
  _cmds[summary_tag]        = summary_cmd;
  _cmds[space_info_tag]     = space_info_cmd;
  _cmds[space_tag]          = space_cmd;
}

