/**
 ** $Id: EventFilters.java 28 2005-06-20 13:13:35Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.interpreter;

import java.util.Arrays;

/**
 * Class for event filter settings.
 * 
 * @author Tony Printezis
 * @author Peter Hofer
 */
public class EventFilters implements Cloneable {

	/** By default a filter is enabled */
	public static final boolean ENABLED_DEFAULT = true;

	/** Default delay */
	public static final int DELAY_DEFAULT = 0;
	/** Minimum delay */
	public static final int DELAY_MIN = 0;
	/** Maximum delay */
	public static final int DELAY_MAX = 10000;

	/** Default is not to pause */
	public static final boolean PAUSE_DEFAULT = false;

	/** Default period */
	public static final int PERIOD_DEFAULT = 1;
	/** Minimum period */
	public static final int PERIOD_MIN = 1;
	/** Maximum period */
	public static final int PERIOD_MAX = 1000;

	private final boolean enabled[];
	private final int delays[]; // in ms
	private final boolean pauses[];
	private final int periods[];

	private final int eventNum;

	/**
	 * Construct a new event filter configuration.
	 * 
	 * @param eventNum
	 *            The number of events
	 * @param enabled
	 *            Which events are enabled
	 * @param delays
	 *            The delays for each event
	 * @param pauses
	 *            The pauses for each event
	 * @param periods
	 *            The periods for each event
	 */
	public EventFilters(int eventNum, boolean enabled[], int delays[], boolean pauses[], int periods[]) {
		this(eventNum, false);
		setEnabled(enabled);
		setDelays(delays);
		setPauses(pauses);
		setPeriods(periods);
	}

	/**
	 * Construct a new event filter configuration with default values.
	 * 
	 * @param eventNum
	 *            The number of events.
	 */
	public EventFilters(int eventNum) {
		this (eventNum, true);
	}

	/**
	 * Private constructor: Creates arrays and optionally initializes them with
	 * default values.
	 * 
	 * @param eventNum
	 *            Number of events.
	 * @param defaults
	 *            <code>true</code> to initialize arrays with default values.
	 */
	private EventFilters(int eventNum, boolean defaults) {
		this.eventNum = eventNum;
		this.enabled = new boolean[eventNum];
		this.delays = new int[eventNum];
		this.pauses = new boolean[eventNum];
		this.periods = new int[eventNum];
		if (defaults) {
			revertToDefaults();
		}
	}

	/**
	 * Set filters for all events to default values.
	 */
	public void revertToDefaults() {
		enableAll();
		clearDelays();
		clearPauses();
		resetPeriods();
	}

	/** Enable all events */
	public void enableAll() {
		Arrays.fill(enabled, true);
	}

	/** Disable all events */
	public void disableAll() {
		Arrays.fill(enabled, false);
	}

	/** Set all delays to default value */
	public void clearDelays() {
		Arrays.fill(delays, DELAY_DEFAULT);
	}

	/** Set all pauses to default value */
	public void clearPauses() {
		Arrays.fill(pauses, PAUSE_DEFAULT);
	}

	/** Set all periods to default value */
	public void resetPeriods() {
		Arrays.fill(periods, PERIOD_DEFAULT);
	}

	/**
	 * Get the number of events
	 * 
	 * @return the number of events
	 */
	public int getNum() {
		return eventNum;
	}

	public boolean[] getEnabled() {
		return enabled;
	}

	public boolean getEnabled(int event) {
		return enabled[event];
	}

	public int[] getDelays() {
		return delays;
	}

	public int getDelay(int event) {
		return delays[event];
	}

	public boolean[] getPauses() {
		return pauses;
	}

	public boolean getPause(int event) {
		return pauses[event];
	}

	public int[] getPeriods() {
		return periods;
	}

	public int getPeriod(int event) {
		return periods[event];
	}

	/**
	 * Copies the event configuration from another {@link EventFilters} instance
	 * to this instance. The number of events must be equal.
	 * 
	 * @param other
	 *            Other {@link EventFilters} object to copy settings from.
	 */
	public void copyFrom(EventFilters other) {
		if (other.eventNum != this.eventNum)
			throw new IllegalArgumentException("Other event filter configuration has different number of events");

		setEnabled(other.enabled);
		setDelays(other.delays);
		setEnabled(other.enabled);
		setPeriods(other.periods);
	}

	public void setEnabled(boolean enabled[]) {
		if (enabled.length != eventNum)
			throw new IllegalArgumentException("array length != eventNum");

		System.arraycopy(enabled, 0, this.enabled, 0, enabled.length);
	}

	public void setEnabled(int event, boolean value) {
		this.enabled[event] = value;
	}

	public void setDelays(int delays[]) {
		if (delays.length != eventNum)
			throw new IllegalArgumentException("array length != eventNum");

		System.arraycopy(delays, 0, this.delays, 0, delays.length);
	}

	public void setDelay(int event, int value) {
		this.delays[event] = value;
	}

	public void setPauses(boolean pauses[]) {
		if (pauses.length != eventNum)
			throw new IllegalArgumentException("array length != eventNum");

		System.arraycopy(pauses, 0, this.pauses, 0, pauses.length);
	}

	public void setPause(int event, boolean value) {
		this.pauses[event] = value;
	}

	public void setPeriods(int periods[]) {
		if (periods.length != eventNum)
			throw new IllegalArgumentException("array length != eventNum");

		System.arraycopy(periods, 0, this.periods, 0, periods.length);
	}

	public void setPeriod(int event, int period) {
		this.periods[event] = period;
	}

	/**
	 * Returns an {@link EventFilters} instance with a copy of this instance's
	 * configuration.
	 */
	@Override
	public Object clone() {
		return new EventFilters(eventNum, enabled, delays, pauses, periods);
	}

}
