package gcspy.tools;

import gcspy.interpreter.Space;
import gcspy.interpreter.Stream;
import gcspy.interpreter.TraceData;
import gcspy.interpreter.client.ClientInterpreter;
import gcspy.interpreter.client.EventCommandListener;
import gcspy.interpreter.client.SpaceCommandListener;
import gcspy.utils.Utils;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

public abstract class OutputStreamTrace extends StreamTrace implements EventCommandListener, SpaceCommandListener {
	private static final Logger logger = Logger.getLogger(OutputStreamTrace.class.getName());
	
	protected abstract ClientInterpreter getInterpreter();
	protected abstract TraceData getData();
	
	private int count, maxCount;

	private long bytesWritten;
	
	public final static int INFINITE = 0;
	
	public int getCount() {
		return count;
	}
	
	public int getMaxCount() {
		return maxCount;
	}

	/******************** For Listeners ********************/

	private boolean shouldReturn() {
		return ((maxCount > 0) && (count >= maxCount));
	}

	public void onEventCmd(int eventID, int elapsedTime, int compensationTime) {
		// catch stray events
		if (shouldReturn()) {
			return;
		}
		
		++count;

		int counts[] = getData().getEventCount();

		if (logger.isLoggable(Level.FINE)) {
			logger.log(Level.FINE, count + ". EVENT  '" + getData().getEvents().getName(eventID) + "'  (" + eventID
					+ ")\n" + "    count = " + counts[eventID] + ",  elapsed = " + elapsedTime + "ms,  compensation = "
					+ compensationTime + "ms");
		}

		try {
			long nextBytesWritten = dumpStreams(eventID, elapsedTime, compensationTime);
			fireEventStoredListeners(eventID, this.bytesWritten, nextBytesWritten - this.bytesWritten);
			this.bytesWritten = nextBytesWritten;
			
			if (shouldReturn()) {
				logger.log(Level.INFO, "Requesting shutdown");
				getInterpreter().sendShutdownReq();
			}
		} catch (IOException e) {
			logger.log(Level.SEVERE, "Exception caught, terminating", e);
			System.exit(-1);
		}
	}

	public void onSpaceCmd(Space space) {
		long nextBytesWritten = dumpSpace(space);
		fireSpaceStoredListeners(space, this.bytesWritten, nextBytesWritten - this.bytesWritten);
		this.bytesWritten = nextBytesWritten;
	}

	private long dumpSpace(Space space) {
		writeSpace(space);
		return bytesWritten();
	}
	private long dumpStreams(int eventID, int elapsedTime, int compensationTime) {		
		int counts[] = getData().getEventCount();
		writeEventCount(counts);

		int spaceNum = getData().getSpaceNum();
		for (int spaceID = 0; spaceID < spaceNum; ++spaceID) {
			Space space = getInterpreter().getSpace(spaceID);
			writeSpaceInfo(spaceID, space.getSpaceInfo());
			
			int streamNum = space.getStreamNum();
			for (int streamID = 0; streamID < streamNum; ++streamID) {
				Stream stream = space.getStream(streamID);
				writeStream(spaceID, streamID, stream.getData(), stream.getSummary());
			}
			
			writeControl(spaceID, space.getControl());
		}

		writeEvent(eventID, elapsedTime, compensationTime);
		
		return bytesWritten();
	}

	/******************** Connection ********************/

	public String generalInfo() {
		Date date = new Date();

		return "Trace Stored By:\n"
				+ "  gcspy.tools.TerminalStoreTrace\n" + "\n"
				+ "Recorded On:\n  " + Utils.formatDateTime(date) + "\n" + "----------\n" + "\n";
	}
	
	private List<EventStoredListener> eventDumpedListeners = new ArrayList<EventStoredListener>();
	
	protected synchronized void addEventStoredListener(EventStoredListener e) {
		assert e != null;
		eventDumpedListeners.add(e);
	}
	
	private synchronized void fireEventStoredListeners(int event, long bytesWrittenSoFar, long size) {
		for (EventStoredListener e : eventDumpedListeners) {
			e.eventStored(event, bytesWrittenSoFar, size);
		}
	}
	
	private synchronized void fireSpaceStoredListeners(Space space, long bytesWrittenSoFar, long size) {
		for (EventStoredListener e : eventDumpedListeners) {
			e.spaceStored(space, bytesWrittenSoFar, size);
		}		
	}

	/******************** Constructor ********************/

	public OutputStreamTrace(OutputStream out, int count, Compression compression) {
		super(out, StreamTrace.DEFAULT_BUFFER_LEN, compression);
		
		this.maxCount = count;

		logger.log(Level.FINE, "Initializing");
	}	
}
