/**
 ** $Id: FileTrace.java 28 2005-06-20 13:13:35Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

/**
 **  gcspy.tools.FileTrace
 **
 **  Contains facilities to store / retrieve traces to / from files
 **/

package gcspy.tools;

import gcspy.comm.BufferedInput;
import gcspy.comm.BufferedOutput;
import gcspy.interpreter.InputGenerator;
import gcspy.interpreter.OutputGenerator;
import gcspy.interpreter.Space;
import gcspy.interpreter.StreamData;
import gcspy.interpreter.TraceHeaderData;
import gcspy.interpreter.StreamData.Datatype;
import gcspy.utils.FileUtils;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

/**
 * Contains facilities to store / retrieve traces to / from Input/OutputStreams
 * 
 * @author Tony Printezis
 * @author Thomas Schatzl
 */
public class StreamTrace implements OutputGenerator, InputGenerator {
	
	private static final Logger logger = Logger.getLogger(StreamTrace.class.getName());

	static private final byte STREAM_CMD = 0;
	static private final byte CONTROL_CMD = 1;
	static private final byte EVENT_COUNT_CMD = 2;
	static private final byte EVENT_CMD = 3;
	static private final byte SPACE_INFO_CMD = 4;
	static private final byte SPACE_CMD = 5;

	public static enum Compression {
		Auto, None, GZip, Zip
	}

	static public final int DEFAULT_BUFFER_LEN = (256 * 1024);

	private StreamTraceOutputStream output = null;
	private StreamTraceInputStream input = null;

	public long bytesWritten() {
		return output.size();
	}

	private Compression compression;

	private byte buffer[];

	/** ****************** Accessor Methods ******************* */

	/**
	 * Get the files compression type (e.g. non, gzip, zip, etc)
	 * 
	 * @return the type
	 */
	protected String getCompressionType() {
		return compression.name();
	}

	/*     ****************** Input/Output ******************* */

	private class StreamTraceOutputStream extends DataOutputStream {

		public StreamTraceOutputStream(OutputStream out) {
			super(out);
		}

	}

	private class StreamTraceInputStream extends DataInputStream {

		public StreamTraceInputStream(InputStream in) {
			super(in);
		}

	}

	/** ****************** Output ******************* */

	/** *** OutputGenerator **** */

	/**
	 * Create a buffered output
	 * 
	 * @return the buffered output object
	 */
	public BufferedOutput createBufferedOutput() throws IOException {
		return new BufferedOutput(buffer);
	}

	/**
	 * Write the given number of bytes from buffer into the output
	 * 
	 * @param len
	 *            the length of the buffer containing data to be written
	 */
	public void writeBuffer(int len) throws IOException {
		output.writeInt(len);
		output.write(buffer, 0, len);
	}

	/**
	 * Flush the output stream
	 */
	public void flush() throws IOException {
		output.flush();
	}
	
	public void writeDataStream(BufferedOutput b, StreamData data) {
		Datatype type = data.getType();
		b.writeInt(type.ordinal());
		switch (type) {
		case Byte:
			b.writeByteArray(data.getBytes());
			break;
		case Short:
			b.writeShortArray(data.getShorts());
			break;
		case Int:
			b.writeIntArray(data.getInts());
			break;
		default:
			throw new Error("Unknown data type");
		}
	}

	public StreamData readStreamData(BufferedInput b) {
		Datatype type = Datatype.map(b.readInt());
		StreamData result = new StreamData(type);
		switch (type) {
		case Byte:
			result.setData(b.readByteArray());
			break;
		case Short:
			result.setData(b.readShortArray());
			break;
		case Int:
			result.setData(b.readIntArray());
			break;
		default:
			throw new Error("Unknown data type");
		}
		return result;
	}

	/** *** STREAM **** */

	/**
	 * Write a stream
	 * 
	 * @param spaceID
	 *            the space's ID
	 * @param streamID
	 *            The stream's ID
	 * @param data
	 *            The stream data
	 * @param summary
	 *            The stream's summary values
	 */
	public void writeStream(int spaceID, int streamID, StreamData data,
			int summary[]) {
		try {
			BufferedOutput b = createBufferedOutput();
			b.writeShort(STREAM_CMD);
			b.writeByte((byte) spaceID);
			b.writeByte((byte) streamID);
			writeDataStream(b, data);
			b.writeIntArray(summary);
			writeBuffer(b.getLen());
		} catch (IOException e) {
			throw new StreamTraceException(e);
		}
	}

	/** *** CONTROL **** */

	/**
	 * Write the controls for a space
	 * 
	 * @param spaceID
	 *            The space's ID
	 * @param control
	 *            The controls
	 */
	public void writeControl(int spaceID, byte control[]) {
		try {
			BufferedOutput b = createBufferedOutput();
			b.writeShort(CONTROL_CMD);
			b.writeByte((byte) spaceID);
			b.writeByteArray(control);
			writeBuffer(b.getLen());
		} catch (IOException e) {
			throw new StreamTraceException(e);
		}
	}

	/** *** EVENT_COUNT **** */

	/**
	 * Write event counts
	 * 
	 * @param counts
	 *            The counts
	 */
	public void writeEventCount(int counts[]) {
		try {
			BufferedOutput b = createBufferedOutput();
			b.writeShort(EVENT_COUNT_CMD);
			b.writeIntArray(counts);
			writeBuffer(b.getLen());
		} catch (IOException e) {
			throw new StreamTraceException(e);
		}
	}

	/** *** EVENT **** */

	/**
	 * Write an event
	 * 
	 * @param eventID
	 *            The event's iD
	 * @param elapsedTime
	 *            The elapsed time for the event
	 * @param compensationTime
	 *            The compensation time
	 */
	public void writeEvent(int eventID, int elapsedTime, int compensationTime) {
		try {
			BufferedOutput b = createBufferedOutput();
			b.writeShort(EVENT_CMD);
			b.writeByte((byte) eventID);
			b.writeInt(elapsedTime);
			b.writeInt(compensationTime);
			writeBuffer(b.getLen());
		} catch (IOException e) {
			throw new StreamTraceException(e);
		}
	}

	/** *** SPACE INFO **** */

	/**
	 * Write the space information for a space
	 * 
	 * @param spaceID
	 *            the space's ID
	 * @param spaceInfo
	 *            the space information
	 */
	public void writeSpaceInfo(int spaceID, String spaceInfo) {
		try {
			BufferedOutput b = createBufferedOutput();
			b.writeShort(SPACE_INFO_CMD);
			b.writeByte((byte) spaceID);		
			if (spaceInfo != null) {
				b.writeString(spaceInfo);
			} else {
				b.writeString("");
			}
			writeBuffer(b.getLen());
		} catch (IOException e) {
			throw new StreamTraceException(e);
		}
	}

	/** *** SPACE **** */

	/**
	 * Write a space
	 * 
	 * @param space
	 *            The space
	 */
	public void writeSpace(Space space) {
		try {
			BufferedOutput buffer = createBufferedOutput();
			buffer.writeShort(SPACE_CMD);
			space.serialise(buffer);
			writeBuffer(buffer.getLen());
		} catch (IOException e) {
			throw new StreamTraceException(e);
		}
	}
	
	public void writeHeader(TraceHeaderData header) {
		try {
			BufferedOutput buffer = createBufferedOutput();
			header.serialise(buffer);
			writeBuffer(buffer.getLen());
		} catch (IOException e) {
			throw new StreamTraceException(e);
		}
	}

	/** ****************** Input ******************* */

	/** *** InputGenerator **** */

	/**
	 * Create a new buffered input
	 * 
	 * @return The new buffered input
	 * @throws IOException
	 */
	public BufferedInput createBufferedInput() throws IOException {
		int len = input.readInt();
		FileUtils.read(input, buffer, 0, len);
		return new BufferedInput(buffer, len);
	}

	/** *** Parsing **** */

	/**
	 * Handle a stream's data and summary values for a stream
	 * 
	 * @param spaceID
	 *            The space's ID
	 * @param streamID
	 *            The stream's ID
	 * @param data
	 *            The stream's data
	 * @param summary
	 *            The summary values
	 * @throws IOException
	 */
	protected void onStreamCmd(int spaceID, int streamID, StreamData data, int summary[])
			throws IOException {
	}

	/**
	 * Handle a space's controls
	 * 
	 * @param spaceID
	 *            The space's ID
	 * @param control
	 *            The controls
	 * @throws IOException
	 */
	protected void onControlCmd(int spaceID, byte control[]) throws IOException {
	}

	/**
	 * Handle event counts
	 * 
	 * @param counts
	 *            The event counts
	 * @throws IOException
	 */
	protected void onEventCountCmd(int counts[]) throws IOException {
	}

	/**
	 * Handle an event
	 * 
	 * @param eventID
	 *            The event's ID
	 * @param elapsedTime
	 *            The elapsed time for the event
	 * @param compensationTime
	 *            Its compensation time
	 * @throws IOException
	 */
	protected void onEventCmd(int eventID, int elapsedTime, int compensationTime)
			throws IOException {
	}

	/**
	 * Handle space information
	 * 
	 * @param spaceID
	 *            The space's ID
	 * @param spaceInfo
	 *            Its space information
	 * @throws IOException
	 */
	protected void onSpaceInfoCmd(int spaceID, String spaceInfo) throws IOException {
	}

	/**
	 * Handle a space
	 * 
	 * @param space
	 *            The space
	 * @throws IOException
	 */
	protected void onSpaceCmd(Space space) throws IOException {
	}

	public void parseOne() throws IOException {
		int spaceID;
		
		BufferedInput b = createBufferedInput();
		
		int tag = b.readShort();
		switch (tag) {
		case STREAM_CMD:
			spaceID = (int) b.readByte();
			int streamID = (int) b.readByte();
			StreamData data = readStreamData(b);
			int summary[] = b.readIntArray();
			onStreamCmd(spaceID, streamID, data, summary);
			break;
		case CONTROL_CMD:
			spaceID = (int) b.readByte();
			byte control[] = b.readByteArray();
			onControlCmd(spaceID, control);
			break;
		case EVENT_COUNT_CMD:
			int counts[] = b.readIntArray();
			onEventCountCmd(counts);
			break;
		case EVENT_CMD:
			int eventID = (int) b.readByte();
			int elapsedTime = b.readInt();
			int compensationTime = b.readInt();
			onEventCmd(eventID, elapsedTime, compensationTime);
			break;
		case SPACE_INFO_CMD:
			spaceID = (int) b.readByte();
			String spaceInfo = b.readString();
			onSpaceInfoCmd(spaceID, spaceInfo);
			break;
		case SPACE_CMD:
			onSpaceCmd(Space.deserialise(b));
			break;
		}
	}

	/**
	 * Parse input
	 * 
	 * @throws IOException
	 */
	public void parse() throws IOException {
        // TODO: Proper loop that doesn't exit by exception only 
		try {
			while (true) {
				parseOne();
			}
		} catch (EOFException ignored) {
		}
	}

	/** ****************** Constructors ******************* */

	public static Compression compressionFor(String filename) {
		String lower = filename.toLowerCase();
		if (lower.endsWith(".gz"))
			return Compression.GZip;
		else if (lower.endsWith(".zip"))
			return Compression.Zip;
		else
			return Compression.None;
	}

	/** Set up the input */
	public void setupInput(InputStream in) {
		try {
			switch (compression) {
			case None:
				input = new StreamTraceInputStream(in);
				break;
			case GZip:
				InputStream cinput = new GZIPInputStream(in);
				input = new StreamTraceInputStream(cinput);
				break;
			case Zip:
				ZipInputStream zipstream = new ZipInputStream(in);
				// only get the first entry
				if (zipstream.getNextEntry() == null) {
					logger.log(Level.SEVERE, "Cannot move to first entry in ZIP input stream, terminating");
					System.exit(-1);
				}
				input = new StreamTraceInputStream(zipstream);
				break;
			default:
				logger.log(Level.SEVERE, "Invalid compression type, terminating");
				System.exit(-1);
			}
		} catch (IOException e) {
			throw new StreamTraceException(e);
		}
	}

	/** Set up the output */
	public void setupOutput(OutputStream out) {

		try {
			switch (compression) {
			case None:
				output = new StreamTraceOutputStream(out);
				break;
			case GZip:
				output = new StreamTraceOutputStream(new GZIPOutputStream(out));
				break;
			case Zip:
				ZipOutputStream zoutput = new ZipOutputStream(out);
				ZipEntry entry = new ZipEntry("default trace");
				zoutput.putNextEntry(entry);
				output = new StreamTraceOutputStream(zoutput);
				break;
			default:
				logger.log(Level.SEVERE, "Invalid compression type, terminating");
				System.exit(-1);
			}
		} catch (IOException e) {
			throw new StreamTraceException(e);
		}
	}

	/** Close the input / output */
	public void close() {
		try {
			if (input != null) {
				input.close();
			}
			if (output != null) {
				output.flush();
				output.close();
			}
		} catch (IOException e) {
			throw new StreamTraceException(e);
		}
	}

	/**
	 * Create a new FileTrace with default buffer length with compression not
	 * set
	 * @param bufferLen
	 *            The length of the buffer
	 * @param compression
	 *            The compression type
	 * @param fileName
	 *            The name of the file
	 */
	public StreamTrace(OutputStream out, int bufferLen, Compression compression) {
		assert compression != Compression.Auto;
		this.compression = compression;
		buffer = new byte[bufferLen];
		setupOutput(out);
	}

	public StreamTrace(InputStream in, int bufferLen, Compression compression) {
		assert compression != Compression.Auto;
		this.compression = compression;
		buffer = new byte[bufferLen];
		setupInput(in);
	}

}
