package gcspy.vis;

import java.awt.Graphics;
import java.awt.Rectangle;

/**
 * Interface for fill patterns for painting.
 * 
 * @author Peter Hofer
 */
public interface FillPattern {

	/**
	 * Fills the intersection of the specified rectangle and the clipping area
	 * of the given {@link Graphics} object with this object's pattern, in the
	 * color set on the {@link Graphics} object.
	 * 
	 * @param g
	 *            The object to use for painting.
	 * @param area
	 *            Rectangle of the area to fill.
	 */
	public void fill(Graphics g, Rectangle area);

	/*  *** Implementations *** */

	/**
	 * Fills the entire clipping area with the color set on the {@link Graphics}
	 * object.
	 */
	public static final FillPattern OPAQUE = new FillPattern() {
		@Override
		public void fill(Graphics g, Rectangle area) {
			Rectangle bounds = area.intersection(g.getClipBounds());
			g.fillRect(bounds.x, bounds.y, bounds.width, bounds.height);
		}
	};

	/**
	 * <pre>
	 * ____
	 * ____
	 * ____
	 * </pre>
	 */
	public static final FillPattern HORIZONTAL_LINES = new FillPattern() {
		private static final int SPACING = 2;
		private static final int THICKNESS = 2;
		private static final int DISTANCE = (SPACING + THICKNESS);

		@Override
		public void fill(Graphics g, Rectangle area) {
			Rectangle bounds = area.intersection(g.getClipBounds());

			int y = bounds.y - (bounds.y % DISTANCE);
			while (y < bounds.y + bounds.height) {
				g.fillRect(bounds.x, y, bounds.width, THICKNESS);
				y += DISTANCE;
			}
		}
	};

	/**
	 * <pre>
	 * ||||
	 * ||||
	 * </pre>
	 */
	public static final FillPattern VERTICAL_LINES = new FillPattern() {
		private static final int SPACING = 2;
		private static final int THICKNESS = 2;
		private static final int DISTANCE = (SPACING + THICKNESS);

		@Override
		public void fill(Graphics g, Rectangle area) {
			Rectangle bounds = area.intersection(g.getClipBounds());

			int x = bounds.x - (bounds.x % DISTANCE);
			while (x < bounds.x + bounds.width) {
				g.fillRect(x, bounds.y, THICKNESS, bounds.height);
				x += DISTANCE;
			}
		}
	};

	/**
	 * <pre>
	 * ////
	 * ////
	 * </pre>
	 */
	public static final FillPattern FORWARD_DIAGONAL_LINES = new FillPattern() {
		private static final int SPACING = 2;
		private static final int THICKNESS = 3;
		private static final int DISTANCE = (SPACING + THICKNESS);

		@Override
		public void fill(Graphics g, Rectangle area) {
			Rectangle bounds = area.intersection(g.getClipBounds());

			int x = bounds.x - bounds.height;
			x = x - (x % DISTANCE) - DISTANCE + (bounds.y % DISTANCE);
			while (x < bounds.x + bounds.width) {
				int[] px = { x, x + THICKNESS, x + THICKNESS + bounds.height, x + bounds.height };
				int[] py = { bounds.y + bounds.height, bounds.y + bounds.height, bounds.y, bounds.y };
				g.fillPolygon(px, py, px.length);
				x += DISTANCE;
			}
		}
	};

	/**
	 * <pre>
	 * \\\\
	 * \\\\
	 * </pre>
	 */
	public static final FillPattern BACKWARD_DIAGONAL_LINES = new FillPattern() {
		private static final int SPACING = 2;
		private static final int THICKNESS = 3;
		private static final int DISTANCE = (SPACING + THICKNESS);

		@Override
		public void fill(Graphics g, Rectangle area) {
			Rectangle bounds = area.intersection(g.getClipBounds());

			int x = bounds.x - bounds.height;
			x = x - (x % DISTANCE) - DISTANCE + (bounds.y % DISTANCE);
			while (x < bounds.x + bounds.width) {
				int[] px = { x, x + THICKNESS, x + THICKNESS + bounds.height, x + bounds.height };
				int[] py = { bounds.y, bounds.y, bounds.y + bounds.height, bounds.y + bounds.height };
				g.fillPolygon(px, py, px.length);
				x += DISTANCE;
			}
		}
	};

	/**
	 * <pre>
	 * +++
	 * +++
	 * </pre>
	 */
	public static final FillPattern GRID = new FillPattern() {
		private static final int SPACING = 3;
		private static final int THICKNESS = 1;
		private static final int DISTANCE = (SPACING + THICKNESS);

		@Override
		public void fill(Graphics g, Rectangle area) {
			Rectangle bounds = area.intersection(g.getClipBounds());

			int x = bounds.x - (bounds.x % DISTANCE);
			while (x < bounds.x + bounds.width) {
				g.fillRect(x, bounds.y, THICKNESS, bounds.height);
				x += DISTANCE;
			}

			int y = bounds.y - (bounds.y % DISTANCE);
			while (y < bounds.y + bounds.height) {
				g.fillRect(bounds.x, y, bounds.width, THICKNESS);
				y += DISTANCE;
			}
		}
	};

	/**
	 * <pre>
	 * /_/_/_
	 * _/_/_/
	 * /_/_/_
	 * </pre>
	 */
	public static final FillPattern GRID_45DEG = new FillPattern() {
		private static final int SPACING = 4;
		private static final int THICKNESS = 2;
		private static final int DISTANCE = (SPACING + THICKNESS);

		@Override
		public void fill(Graphics g, Rectangle area) {
			Rectangle bounds = area.intersection(g.getClipBounds());
			
			int x = bounds.x - bounds.height;
			x = x - (x % DISTANCE) - DISTANCE + (bounds.y % DISTANCE);
			while (x < bounds.x + bounds.width) {
				int[] px = { x, x + THICKNESS, x + THICKNESS + bounds.height, x + bounds.height };
				int[] py1 = { bounds.y + bounds.height, bounds.y + bounds.height, bounds.y, bounds.y };
				int[] py2 = { bounds.y, bounds.y, bounds.y + bounds.height, bounds.y + bounds.height };
				g.fillPolygon(px, py1, px.length);
				g.fillPolygon(px, py2, px.length);
				x += DISTANCE;
			}
		}
	};
	
	/**
	 * <pre>
	 * . . .
	 * . . .
	 * . . .
	 * </pre>
	 */
	public static final FillPattern DOTS = new FillPattern() {
		private static final int SPACING = 4;
		private static final int DIAMETER = 3;
		private static final int DISTANCE = (SPACING + DIAMETER);

		@Override
		public void fill(Graphics g, Rectangle area) {
			Rectangle bounds = area.intersection(g.getClipBounds());

			int y = bounds.y - (bounds.y % DISTANCE);
			while (y < bounds.y + bounds.height) {
				int x = bounds.x - (bounds.x % DISTANCE);
				while (x < bounds.x + bounds.width) {
					g.fillOval(x, y, DIAMETER, DIAMETER);
					x += SPACING;
				}
				y += SPACING;
			}
		}
	};

	public static final FillPattern[] ALL_PATTERNS = {
		FillPattern.FORWARD_DIAGONAL_LINES,
		FillPattern.HORIZONTAL_LINES,
		FillPattern.DOTS,
		FillPattern.GRID_45DEG,
		FillPattern.GRID,
		FillPattern.BACKWARD_DIAGONAL_LINES,
		FillPattern.VERTICAL_LINES,
		FillPattern.OPAQUE
	};
}
