/**
 ** $Id: MainFrame.java 41 2006-08-15 14:26:59Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.vis;

import gcspy.comm.SocketServer;
import gcspy.interpreter.SingleEventTraceData;
import gcspy.interpreter.Space;
import gcspy.interpreter.TraceData;
import gcspy.interpreter.client.ClientInterpreter;
import gcspy.interpreter.client.EventCommandListener;
import gcspy.interpreter.client.PauseListener;
import gcspy.tools.EventInterpreter;
import gcspy.tools.StreamTrace;
import gcspy.tools.TerminalPlayTrace;
import gcspy.tools.StreamTrace.Compression;
import gcspy.utils.FileUtils;
import gcspy.vis.ConnectionDialog.ConnectionType;
import gcspy.vis.ConnectionInfo.Status;
import gcspy.vis.plugins.PluginManager;
import gcspy.vis.utils.Factory;
import gcspy.vis.utils.IconLoader;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dialog.ModalityType;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.GZIPOutputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;

import bibliothek.gui.dock.common.CControl;
import bibliothek.gui.dock.common.CGrid;
import bibliothek.gui.dock.common.DefaultSingleCDockable;
import bibliothek.gui.dock.common.SingleCDockable;
import bibliothek.gui.dock.common.action.CAction;
import bibliothek.gui.dock.common.action.CButton;
import bibliothek.gui.dock.common.intern.CDockable;
import bibliothek.gui.dock.common.layout.ThemeMap;

/**
 * Main frame of the visualiser
 * 
 * @author Tony Printezis
 * @author <a href="http://www/cs/kent.ac.uk">Richard Jones</a>
 * @author Thomas Schatzl
 * @author Peter Hofer
 */
public class MainFrame extends JFrame implements
		EventCommandListener, PauseListener {

	private static final String DEFAULT_DOCKING_THEME = ThemeMap.KEY_ECLIPSE_THEME;
	private static final String DOCKABLE_SPACE_PREFIX = "space__";

	private static final Logger logger = Logger.getLogger(MainFrame.class.getName());

	private CControl dockController;

	private JButton connectB;
	private boolean connected;

	private JTextField activityTF;
	private ConnectionInfo connectedL;

	private JButton backB;
	private JButton pauseB;
	private JButton playB;
	private JButton playOneB;
	private EventSlider eventS;

	private JPanel dockP;

	private JPanel bottomP;

	// Menu
	private JMenu traceM;
	private JMenuItem saveTraceMI;
	private JMenu utilitiesM;
	private JCheckBoxMenuItem propertiesMI;
	private JCheckBoxMenuItem generalInfoMI;
	private JCheckBoxMenuItem timersMI;
	private JCheckBoxMenuItem eventCountMI;
	private JMenuItem eventFilterMI;
	private JMenu spacesM;
	private JMenu pluginsM;

	private UIInterpreterProxy proxy;
	private ClientInterpreter interpreter;
	private PlayTraceThread playTraceThread;
	private TraceData interpreterData;

	private String host = null;
	private int port = -1;
	private boolean pauseAtStart = false;

	/**
	 * The name of the trace file used for playing data
	 */
	private String traceFile;

	private boolean connectOpenFile;

	/**
	 * Default index file name used when failing to create a temporary file name.
	 */
	public static final File DEFAULT_INDEX_FILE = new File(System.getProperty("java.io.tmpdir"), "gcspy-index");
	
	/**
	 * The name of the index file. It is used to store the trace data received
	 * so far to implement backward/forward skipping
	 */
	private File indexFile = DEFAULT_INDEX_FILE;

	private boolean doConnect = false;

	private int curEvent;

	private int getNumEvents() {
		int count = 0;
		if (indexer != null) {
			count = indexer.getNumEvents();
		}
		return count;
	}

	private ClientInterpreterThread mainLoopThread;

	private SingleCDockable propertiesDockable;
	private TilePropertyView propertiesView;

	private SingleCDockable eventCountersDockable;
	private EventCountView eventCountView;

	private SingleCDockable generalInfoDockable;
	private JTextArea generalInfoView;

	private SingleCDockable timersDockable;
	private TimersView timersView;

	private List<DisconnectListener> disconnectListeners = new ArrayList<DisconnectListener>();

	private PluginManager pluginManager;

	private Indexer indexer;

	/* *** Listeners *** */

	/**
	 * Add a disconnect listener
	 * 
	 * @param listener
	 *            the listener
	 */
	public void addDisconnectListener(DisconnectListener listener) {
		disconnectListeners.add(listener);
	}

	/**
	 * Invokes the registered listeners for the "Disconnect" event. If
	 * reconnecting is set, automatically clears the list.
	 * 
	 * @param reconnecting
	 *            Set if the event is a reconnect, e.g. if a connection attempt
	 *            will be tried soon.
	 */
	private void fireDisconnectListeners(boolean reconnecting) {
		for (DisconnectListener listener : disconnectListeners)
			listener.disconnect(reconnecting);
		if (reconnecting) {
			Iterator<DisconnectListener> it = disconnectListeners.iterator();
			while (it.hasNext()) {
				it.next();
				it.remove();
			}
		}
	}

	private ActionListener saveToFileMListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JFileChooser fc = new JFileChooser();
			int response = fc.showSaveDialog(MainFrame.this);
			if (response == JFileChooser.APPROVE_OPTION) {
				saveIndexToFile(fc.getSelectedFile());
			} else {
				// canceled by user
			}
		}
	};

	private ActionListener connectBListener = new ActionListener() {
		public void actionPerformed(ActionEvent event) {
			if (!connected) {
				ConnectionDialog cd = new ConnectionDialog(MainFrame.this, host, port, pauseAtStart, false,
						ConnectionDialog.DEFAULT_TRACE_FILE);
				cd.setLocationRelativeTo(connectB);
				cd.setVisible(true);
				if (cd.getDialogResult() == ConnectionDialog.RESULT_OK) {
					host = cd.getHost();
					port = cd.getPort();
					pauseAtStart = cd.getPauseAtStart();
					traceFile = cd.getTraceFile();
					connectOpenFile = (cd.getConnectionType() == ConnectionType.FILE);

					curEvent = 0;

					connect(host, port, pauseAtStart);
					eventS.setEnabled(pauseAtStart);
				} else {
					setConnected(false);
				}
			} else {
				disconnect();
			}
		}
	};

	private void setConnected(boolean connected) {
		if (this.connected != connected) {
			this.connected = connected;
			if (connected) {
				connectB.setText("Disconnect");
				connectedL.setStatus(Status.Connected);
			} else {
				connectB.setText("Connect...");
				connectedL.setStatus(Status.Disconnected);
			}
		}
	}

	private ActionListener eventNavigationListener = new ActionListener() {
		public void actionPerformed(ActionEvent event) {
			Object source = event.getSource();
			if (source == backB) {
				int nextEvent = Math.max(0, curEvent - 1);
				if (nextEvent != curEvent) {
					eventS.moveSlider(nextEvent);
				}
			} else if (source == pauseB) {
				pauseReq();
			} else if (source == playB) {
				play();
			} else if (source == playOneB) {
				int nextEvent = Math.min(getNumEvents(), curEvent + 1);
				if (nextEvent != curEvent) {
					eventS.moveSlider(nextEvent);
				}
			} else {
				throw new RuntimeException("Received action event from unknown source");
			}
		}
	};

	private WindowAdapter mainWindowListener = new WindowAdapter() {
		@Override
		public void windowOpened(WindowEvent e) {
			if (doConnect) {
				connect(host, port, pauseAtStart);
			} else {
				connectB.doClick();
			}
		}

		@Override
		public void windowClosing(WindowEvent e) {
			quit();
		}
	};

	private ActionListener quitMIListener = new ActionListener() {
		public void actionPerformed(ActionEvent event) {
			quit();
		}
	};

	private ActionListener aboutMIListener = new ActionListener() {
		public void actionPerformed(ActionEvent event) {
			JOptionPane.showMessageDialog(MainFrame.this,
				"GCspy Heap Visualization Framework\n\n" +
				"Tony Printezis\n" +
				"Richard Jones\n" +
				"Thomas Schatzl\n" +
				"Peter Hofer", "About GCspy",
				JOptionPane.INFORMATION_MESSAGE);
		}
	};

	private ActionListener utilitiesMIListener = new ActionListener() {
		public void actionPerformed(ActionEvent event) {
			Object source = event.getSource();
			if (source == generalInfoMI) {
				generalInfoDockable.setVisible(!generalInfoDockable.isVisible());
			} else if (source == timersMI) {
				timersDockable.setVisible(!timersDockable.isVisible());
			} else if (source == eventCountMI) {
				eventCountersDockable.setVisible(!eventCountersDockable.isVisible());
			} else if (source == eventFilterMI) {
				EventFilterDialog dialog = new EventFilterDialog(MainFrame.this,
						interpreterData.getEvents().getNames(), interpreterData.getEventFilters());
				dialog.setModalityType(ModalityType.APPLICATION_MODAL);
				dialog.setVisible(true);
				if (dialog.getDialogResult() == EventFilterDialog.RESULT_OK) {
					interpreterData.setEventFilters(dialog.getEventFilters());
					try {
						interpreter.sendEventFilters();
					} catch (IOException e) {
						JOptionPane.showMessageDialog(MainFrame.this,
							"Event filter configuration could not be applied.", "Event Filters",
							JOptionPane.WARNING_MESSAGE);
						logger.log(Level.WARNING, "Sending event filters failed", e);
					}
				}
				dialog.dispose();
			} else if (source == propertiesMI) {
				propertiesDockable.setVisible(!propertiesDockable.isVisible());
			} else {
				throw new RuntimeException("Received action event from unknown source");
			}
		}
	};

	private ActionListener pluginMIListener = new ActionListener() {
		public void actionPerformed(ActionEvent e) {
			JCheckBoxMenuItem cbmi = (JCheckBoxMenuItem) e.getSource();
			String plugin = cbmi.getText();
			boolean enabled = cbmi.getState();
			pluginManager.setActivePlugin(plugin, enabled);
		}
	};

	private EventSliderMovedListener eventSliderListener = new EventSliderMovedListener() {
		@Override
		public void eventSliderMoved(int event) {
			if (indexer != null) {
				if (event != curEvent && event >= 0 && event < indexer.getNumEvents()) {
					playOne(event);
				}
			}
		}
	};

	private EnhancedThreadListener threadListener = new EnhancedThreadListener() {
		@Override
		public void threadCompleted(Thread thread) {
			if (thread == mainLoopThread) {
				logger.log(Level.INFO, "Main loop thread completed");

				SwingUtilities.invokeLater(new Runnable() {
					@Override
					public void run() {
						mainLoopThread = null;
						interpreter = null;

						setSynchronous(false); // pause
					}
				});

			} else if (thread == playTraceThread) {
				logger.log(Level.INFO, "Play trace thread completed");

				SwingUtilities.invokeLater(new Runnable() {
					@Override
					public void run() {
						playTraceThread = null;
					}
				});
			}
		}
	};

	public void saveIndexToFile(File file) {
		Compression comp = StreamTrace.compressionFor(file.getName());

		try {
			InputStream in = new FileInputStream(indexFile);
			OutputStream out = new FileOutputStream(file);
	
			switch (comp) {
			case GZip:
				out = new GZIPOutputStream(out);
				break;
			case Zip:
				ZipOutputStream zipout = new ZipOutputStream(out);
				ZipEntry entry = new ZipEntry("trace");
				zipout.putNextEntry(entry);
				out = zipout;
				break;
			case None:
				break;
			default:
				throw new IllegalArgumentException("Could not determine compression for file");
			}

			FileUtils.copy(in, out);

			in.close();
			out.close();
		} catch (IOException e) {
			logger.log(Level.WARNING, "Exception caught", e);
		}
	}

	private void setCurrentEvent(String text) {
		activityTF.setText("Current Event: " + text);
	}

	/* *** Some Events **** */

	/**
	 * Request shutdown and wait for the main loop to end.
	 */
	private void requestShutdown() {
		if (interpreter != null) {
			try {
				connectedL.setStatus(Status.Disconnecting);
				interpreter.sendShutdownReq();

				while (mainLoopThread != null && mainLoopThread.isAlive()) {
					try {
						mainLoopThread.join();
					} catch (InterruptedException e) {
						// ignore
					}
				}
			} catch (IOException e) {
				logger.log(Level.SEVERE, "Exception caught, terminating", e);
			}
		}
	}

	private void quit() {
		disconnect();
		System.exit(0);
	}

	private void connect(String host, int port, boolean pauseAtStart) {
		connectedL.setEndpoint(host + ":" + port);
		connectedL.setStatus(Status.Connecting);
		fireDisconnectListeners(true);

		try {
			interpreterData = new TraceData();

			// When opening a file, start a thread that acts as server and
			// replays the contents of the file
			if (connectOpenFile) {
				port = SocketServer.ANY_PORT;

				TerminalPlayTrace playTrace = new TerminalPlayTrace(traceFile);
				playTraceThread = new PlayTraceThread(playTrace);
				playTraceThread.addCompletionListener(threadListener);
				playTraceThread.start();

				do {
					port = playTrace.getBindPort();
					Thread.sleep(10);
				} while (port == SocketServer.ANY_PORT);
			}

			// Create interpreter
			interpreter = new ClientInterpreter(interpreterData);

			// Set up the indexer
			indexer = null;
			try {
				// get a new temporary file name for the index data. If this is not possible, try the default
				// index file, but not before issuing a warning...
				try {
					indexFile = File.createTempFile("gcspy-index", null);
					logger.log(Level.INFO, "Using file \"" + indexFile.getName() + "\" as index.");
				} catch (IOException e) {
					indexFile = DEFAULT_INDEX_FILE;
					logger.log(Level.SEVERE, "Could not assign a new temporary index file. Trying to use default file called \"" +
					    DEFAULT_INDEX_FILE.getName() + "\". Running multiple instances of this program may fail.");
				}			
				
				if (indexFile.createNewFile() || indexFile.canWrite()) {
					indexer = new Indexer(interpreter, interpreterData, new FileOutputStream(indexFile));
				} else {
					logger.log(Level.WARNING, "Cannot write indexing file, disabling related functionality (seeking).");
				}
			} catch (IOException e) {
				logger.log(Level.WARNING, "Exception caught", e);
			}

			// Set up event handling
			proxy = new UIInterpreterProxy(interpreterData);
			proxy.addEventListener(this);
			proxy.addPauseListener(this);
			interpreter.addPauseListener(proxy);
			interpreter.addEventListener(proxy);
			interpreter.addSpaceListener(proxy);

			// Connect
			interpreter.connectToServer(host, port, pauseAtStart);

			// Set up user interface
			setupSpaces();
			saveTraceMI.setEnabled(true);
			eventCountView.setTrace(interpreterData, proxy);
			generalInfoView.setText(interpreterData.getGeneralInfo());
			timersView.setEventProvider(proxy);
			setMenuItemsEnabled(utilitiesM, true);
			setMenuItemsEnabled(pluginsM, false);

			pauseB.setEnabled(!pauseAtStart);

			mainLoopThread = new ClientInterpreterThread(interpreter);
			mainLoopThread.addCompletionListener(threadListener);
			mainLoopThread.start();

			connectedL.setName(interpreter.getName());
			setConnected(true);
		} catch (Exception e) {
			JOptionPane.showMessageDialog(this, "Connection failed:\n\n" + e.getLocalizedMessage(),
				"Connection failed", JOptionPane.WARNING_MESSAGE);
			disconnect();
		}
	}

	private void disconnect() {
		try {
			requestShutdown();
		} catch (Exception e) {
			logger.log(Level.WARNING, "Exception when disconnecting (ignoring)", e);
		}
		
		setConnected(false);

		if (indexer != null) {
			try {
				indexer.close();
			} catch (Exception e) {
				logger.log(Level.WARNING, "Exception when closing indexer (ignored)", e);
			}
			indexer = null;
			indexFile.delete();
		}
		interpreterData = null;

		setCurrentEvent("-");
		setNavigationEnabled(false);
		setMenuItemsEnabled(utilitiesM, false);
		setMenuItemsEnabled(pluginsM, true);
		eventS.reset();
		saveTraceMI.setEnabled(false);
		clearDockingArea();
		spacesM.removeAll();

		fireDisconnectListeners(false);
	}

	public void onEventCmd(int event, int elapsedTime, int compensationTime) {
		if (interpreterData != null) { // test for late events after disconnect
			String eventName = interpreterData.getEvents().getName(event);
			setCurrentEvent(eventName + " (" + curEvent + "/" + getNumEvents() + ")");
		}
	}

	private void pauseReq() {
		try {
			if (interpreter != null) {
				connectedL.setStatus(Status.Pausing);
				interpreter.sendPauseReq();
				connectedL.setStatus(Status.Paused);
			} else {
				setSynchronous(false);
			}
		} catch (Exception e) {
			logger.log(Level.SEVERE, "Exception caught, terminating", e);
		}
	}

	private void play() {
		try {
			if (interpreter != null) {
				setSynchronous(true);
				interpreter.sendRestart();
				connectedL.setStatus(Status.Connected);
			} else {
				setSynchronous(false);
			}
		} catch (Exception e) {
			logger.log(Level.SEVERE, "Exception caught, terminating", e);
		}
	}

	private void setSynchronous(boolean synchronous) {
		proxy.setSynchronous(synchronous);

		// enable back, forward etc. only if an indexer is available
		setNavigationEnabled(!synchronous && indexer != null);

		// always enable either pause or play (when actively connected)
		pauseB.setEnabled(synchronous && interpreter != null);
		playB.setEnabled(!synchronous && interpreter != null);
	}

	private void setNavigationEnabled(boolean enabled) {
		backB.setEnabled(enabled);
		pauseB.setEnabled(enabled);
		playB.setEnabled(enabled);
		playOneB.setEnabled(enabled);
		eventS.setEnabled(enabled);
	}

	public void onPause() {
		setSynchronous(false);
	}

	/**
	 * Enables or disables all items of a menu.
	 */
	private void setMenuItemsEnabled(JMenu menu, boolean enabled) {
		Component[] components = menu.getMenuComponents();
		for (Component c : components) {
			c.setEnabled(enabled);
		}
	}
	
	/* *** Setup *** */

	/**
	 * Removes all spaces and default dockables from the docking area and sets
	 * the associated member variables to <code>null</code>.
	 */
	private void clearDockingArea() {
		// NOTE: List changes when remove() is called on the controller,
		// so it has to be copied
		List<SingleCDockable> list = new ArrayList<SingleCDockable>(dockController.getRegister().getSingleDockables());
		for (SingleCDockable dock : list) {
			dockController.remove(dock);
		}
		
		dockP.setVisible(false);

		propertiesView = null;
		propertiesDockable = null;

		generalInfoView = null;
		generalInfoDockable = null;

		eventCountView = null;
		eventCountersDockable = null;

		timersView = null;
		timersDockable = null;
	}

	/**
	 * Setup up the docking area, creating the default dockables and a
	 * {@link SpaceManager}, a dockable and a menu entry for each space.
	 */
	private void setupSpaces() {
		CGrid grid = new CGrid(dockController);
		
		setupDefaultDockables();

		// Add "general-purpose" dockables
		grid.add(2, 0, 1, 1, propertiesDockable);
		grid.add(2, 1, 1, 1, generalInfoDockable);
		grid.add(2, 2, 1, 1, eventCountersDockable);
		grid.add(2, 3, 1, 1, timersDockable);
		
		// Dockables can only be hidden after they are added to the CControl
		// (which happens when deploying the layout), so keep a list of
		// dockables to hide later on.
		List<CDockable> hideLater = new ArrayList<CDockable>(
				Arrays.asList(generalInfoDockable, eventCountersDockable, timersDockable));

		// Add one dockable for each space
		int nspaces = interpreterData.getSpaceNum();
		for (int i = 0; i < nspaces; ++i) {
			Space space = interpreterData.getSpace(i);
			SpaceManager spaceManager = new SpaceManager(this, proxy, space, pluginManager);
			addDisconnectListener(spaceManager);
			spaceManager.setPropertyView(propertiesView);

			final DefaultSingleCDockable dockable = new DefaultSingleCDockable(
				DOCKABLE_SPACE_PREFIX + space.getID(),
				space.getFullName(),
				spaceManager);
			dockable.setCloseable(true);
			grid.add(0, i, 2, 1, dockable);

			// Summary and Legend frame
			CDockable summary = spaceManager.getSummaryDockable();
			CDockable legend = spaceManager.getLegendDockable();
			grid.add(2, i + 4, 1, 0.5, legend);
			grid.add(2, i + 4.5, 1, 0.5, summary);
			hideLater.add(summary);
			hideLater.add(legend);

			// Insert menu item
			JCheckBoxMenuItem item = new JCheckBoxMenuItem(space.getFullName());
			item.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					dockable.setVisible(!dockable.isVisible());
				}
			});
			spacesM.add(item);

			dockable.addCDockableStateListener(new ButtonSelectDockableListener(item));
		}

		dockController.getContentArea().deploy(grid);

		// Hide dockables now that they are registered
		for (CDockable dock : hideLater) {
			dock.setVisible(false);
		}

		interpreter.addEventListener(eventS);
		proxy.setListenersEnabled(true);

		dockP.setVisible(true);
	}

	/**
	 * Creates instances for the default dockables (properties etc.) and the
	 * views (components) they contain.
	 */
	private void setupDefaultDockables() {
		propertiesView = new TilePropertyView();
		CAction copyAction = new CButton("Copy to Clipboard", IconLoader.getIcon(IconLoader.COPY_ICON)) {
			@Override
			protected void action() {
				propertiesView.copyToClipboard();
			}
		};
		DefaultSingleCDockable pdock = new DefaultSingleCDockable("$properties$", "Properties", propertiesView, copyAction);
		pdock.setCloseable(true);
		pdock.addCDockableStateListener(new ButtonSelectDockableListener(propertiesMI));
		propertiesDockable = pdock;

		generalInfoView = Factory.createTextArea(false, true);
		DefaultSingleCDockable gidock = new DefaultSingleCDockable("$generalinfo$", "General Info", generalInfoView);
		gidock.setCloseable(true);
		gidock.addCDockableStateListener(new ButtonSelectDockableListener(generalInfoMI));
		generalInfoDockable = gidock;

		eventCountView = new EventCountView();
		DefaultSingleCDockable ecdock = new DefaultSingleCDockable("$eventcounters$", "Event Counters", eventCountView);
		ecdock.setCloseable(true);
		ecdock.addCDockableStateListener(new ButtonSelectDockableListener(eventCountMI));
		eventCountersDockable = ecdock;

		timersView = new TimersView();
		DefaultSingleCDockable tdock = new DefaultSingleCDockable("$timers$", "Timers", timersView);
		tdock.setCloseable(true);
		tdock.addCDockableStateListener(new ButtonSelectDockableListener(timersMI));
		timersDockable = tdock;
	}

	private void setupBottomPanel() {
		Container horizontal;

		bottomP = Factory.createBorderPanel();

		eventS = new EventSlider();
		eventS.setEnabled(false);
		eventS.addMovedListener(eventSliderListener);
		bottomP.add(eventS, BorderLayout.CENTER);

		horizontal = new Box(BoxLayout.X_AXIS);
		connectB = Factory.createButton("Connect...", connectBListener);
		setConnected(doConnect);
		horizontal.add(connectB);
		bottomP.add(horizontal, BorderLayout.EAST);

		horizontal = new Box(BoxLayout.X_AXIS);
		backB = Factory.createIconButton(IconLoader.getIcon(IconLoader.BACK_ONE_ICON), false,
				"Back one step", eventNavigationListener);
		horizontal.add(backB);
		pauseB = Factory.createIconButton(IconLoader.getIcon(IconLoader.PAUSE_ICON), false,
				"Pause", eventNavigationListener);
		horizontal.add(pauseB);
		playB = Factory.createIconButton(IconLoader.getIcon(IconLoader.PLAY_ICON), false,
				"Play", eventNavigationListener);
		horizontal.add(playB);
		playOneB = Factory.createIconButton(IconLoader.getIcon(IconLoader.PLAY_ONE_ICON),
				false, "Step One", eventNavigationListener);
		horizontal.add(playOneB);
		bottomP.add(horizontal, BorderLayout.WEST);

		JPanel statusP = Factory.createBorderPanel();
		statusP.setBorder(BorderFactory.createEtchedBorder());
		bottomP.add(statusP, BorderLayout.SOUTH);
		activityTF = Factory.createTextField(20, false);
		activityTF.setBorder(BorderFactory.createEmptyBorder());
		statusP.add(activityTF, BorderLayout.CENTER);
		connectedL = new ConnectionInfo();
		statusP.add(connectedL, BorderLayout.EAST);
	}

	private void setupMenu() {
		JMenuBar menuBar = new JMenuBar();

		traceM = new JMenu("Trace");
		traceM.setMnemonic(KeyEvent.VK_T);
		menuBar.add(traceM);

		saveTraceMI = new JMenuItem("Save Trace As...", KeyEvent.VK_S);
		saveTraceMI.addActionListener(saveToFileMListener);
		saveTraceMI.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, InputEvent.CTRL_DOWN_MASK));
		traceM.add(saveTraceMI);

		traceM.addSeparator();

		JMenuItem exit = new JMenuItem("Exit", KeyEvent.VK_X);
		exit.addActionListener(quitMIListener);
		exit.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, InputEvent.CTRL_DOWN_MASK));
		traceM.add(exit);

		utilitiesM = new JMenu("Utilities");
		utilitiesM.setMnemonic(KeyEvent.VK_U);
		menuBar.add(utilitiesM);

		propertiesMI = new JCheckBoxMenuItem("Properties");
		propertiesMI.setMnemonic(KeyEvent.VK_P);
		propertiesMI.addActionListener(utilitiesMIListener);
		utilitiesM.add(propertiesMI);

		generalInfoMI = new JCheckBoxMenuItem("General Info");
		propertiesMI.setMnemonic(KeyEvent.VK_G);
		generalInfoMI.addActionListener(utilitiesMIListener);
		utilitiesM.add(generalInfoMI);

		timersMI = new JCheckBoxMenuItem("Timers");
		propertiesMI.setMnemonic(KeyEvent.VK_T);
		timersMI.addActionListener(utilitiesMIListener);
		utilitiesM.add(timersMI);

		eventCountMI = new JCheckBoxMenuItem("Event Counters");
		propertiesMI.setMnemonic(KeyEvent.VK_C);
		eventCountMI.addActionListener(utilitiesMIListener);
		utilitiesM.add(eventCountMI);

		eventFilterMI = new JMenuItem("Event Filters...", KeyEvent.VK_F);
		eventFilterMI.addActionListener(utilitiesMIListener);
		utilitiesM.add(eventFilterMI);

		spacesM = new JMenu("Spaces");
		spacesM.setMnemonic(KeyEvent.VK_S);
		menuBar.add(spacesM);

		pluginsM = new JMenu("Plugins");
		pluginsM.setMnemonic(KeyEvent.VK_P);
		menuBar.add(pluginsM);
		
		Set<String> allPlugins = pluginManager.getAllPlugins();
		for (String plugin : allPlugins) {
			boolean selected = pluginManager.isActivePlugin(plugin);
			pluginsM.add(Factory.createCheckBoxMenuItem(plugin, selected, pluginMIListener));
		}

		JMenu helpMenu = new JMenu("Help");
		helpMenu.setMnemonic(KeyEvent.VK_H);
		menuBar.add(helpMenu);

		JMenuItem about = new JMenuItem("About", KeyEvent.VK_A);
		about.addActionListener(aboutMIListener);
		helpMenu.add(about);

		setJMenuBar(menuBar);
	}

	private void setupComponents() {	
		Container container = getContentPane();
		container.setLayout(new BorderLayout());

		// Main docking panel
		dockP = Factory.createBorderPanel();
		// Setup up controller
		dockController = new CControl(this);
		dockController.setTheme(DEFAULT_DOCKING_THEME);
		dockP.add(dockController.getContentArea(), BorderLayout.CENTER);
		container.add(dockP, BorderLayout.CENTER);

		// Bottom panel
		setupBottomPanel();
		container.add(bottomP, BorderLayout.SOUTH);

		// Menu
		setupMenu();
	}

	private void playOne(int nextEvent) {
		if (indexer == null)
			return;
		
		// Proxy is configured not to relay events
		if (!proxy.isListenersEnabled())
			return;

		FileInputStream f = null;
		try {
			f = new FileInputStream(indexFile);

			long framePos = indexer.getFramePosAt(nextEvent);
			FileUtils.seek(f, framePos);

			// read the data for the complete event
			int size = (int) indexer.getFrameSizeAt(nextEvent);
			byte[] data = new byte[size];
			FileUtils.read(f, data, 0, data.length);

			ByteArrayInputStream in = new ByteArrayInputStream(data);
			// issue event

			SingleEventTraceData trace = new SingleEventTraceData(interpreterData.getSpaceNum());
			for (int i = 0; i < indexer.getNumSpaces(); i++) {
				trace.setSpace(i, indexer.getSpace(i));
			}

			EventInterpreter interpreter = new EventInterpreter(in, trace);
			interpreter.execute();

			for (int i = 0; i < indexer.getNumSpaces(); i++) {
				proxy.fireSpaceListeners(indexer.getSpace(i));
			}

			proxy.fireEventListeners(
				interpreter.getLastEventId(),
				interpreter.getLastElapsedTime(),
				interpreter.getLastCompensationTime());

			// display
			curEvent = nextEvent;
		} catch (IOException e) {
			logger.log(Level.WARNING, "Exception caught", e);
		} finally {
			try {
				f.close();
			} catch (IOException e) {
				// ignore
			}
		}
	}

	/* *** Constructor **** */

	/**
	 * Construct a new MainFrame
	 * 
	 * @param host
	 *            the server's hostname
	 * @param port
	 *            The port on which to connect
	 * @param pauseAtStart
	 *            true if the server is to pause at the start
	 * @param doConnect
	 *            true if the visualizer is to connect
	 */
	public MainFrame(String host, int port, boolean pauseAtStart, boolean doConnect) {
		super("GCspy");
		addWindowListener(mainWindowListener);

		this.host = host;
		this.port = port;
		this.pauseAtStart = pauseAtStart;
		this.doConnect = doConnect;

		pluginManager = new PluginManager();
		setupComponents();

		// Initialize to disconnected state
		disconnect();

		pack();
	}

}
