package gcspy.vis;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.StringSelection;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.ListCellRenderer;
import javax.swing.ListModel;
import javax.swing.ListSelectionModel;
import javax.swing.SwingConstants;

import net.miginfocom.swing.MigLayout;

/**
 * Component which, provided with a {@link ListModel}, displays
 * {@link PropertyItem} objects consisting of name, value and an optional bar
 * chart.
 * 
 * @author Peter Hofer
 */
public class PropertyListView extends JComponent {

	/**
	 * Component that draws a diagram bar for which the length relative to the
	 * component's width depends on the displayed percentage.
	 * 
	 * @author Peter Hofer
	 */
	private static class DiagramBar extends JLabel {

		public static final Color DEFAULT_FOREGROUND = Color.BLUE;
		public static final Color DEFAULT_BACKGROUND = Color.WHITE;

		private double value;

		public DiagramBar() {
			this(0, DEFAULT_FOREGROUND, DEFAULT_BACKGROUND);
		}

		public DiagramBar(double percentage, Color foreground, Color background) {
			setPercentage(percentage);

			setForeground(foreground);
			setBackground(background);

			setBorder(BorderFactory.createEtchedBorder());
		}

		public void setPercentage(double percentage) {
			percentage = Math.max(0, Math.min(percentage, 100));
			this.value = percentage / 100.0;
		}

		@Override
		protected void paintComponent(Graphics g) {
			Graphics2D g2d = (Graphics2D) g;
			g2d.setColor(getBackground());
			g2d.fill(g2d.getClip());
			g2d.setColor(getForeground());
			g2d.fillRect(0, 0, (int) (getWidth() * value), getHeight());
		}
	}

	/**
	 * Component that displays a {@link PropertyItem} showing the name on the
	 * left, a bar displaying the percentage in the center (if enabled) and the
	 * value on the right.
	 * 
	 * @author Peter Hofer
	 */
	private static class PropertyItemComponent extends JComponent {

		protected JLabel name = new JLabel();
		protected JLabel value = new JLabel("", SwingConstants.RIGHT);
		private DiagramBar bar = new DiagramBar();
		
		private final Font NORMAL_FONT = name.getFont();
		private final Font HEADING_FONT = name.getFont().deriveFont(Font.BOLD);

		PropertyItemComponent() {
			setOpaque(true);
			setLayout(new MigLayout("fill, insets 1"));
			add(name, "wmin 35%");
			add(bar, "grow, push");
			add(value, "wmin 15%");
		}

		public void setNameString(String newName) {
			name.setText(newName);
		}

		public void setValueString(String newValue) {
			value.setText(newValue);
		}

		public void setBarColor(Color color) {
			bar.setForeground(color);
		}

		public void setBarPercentage(double percentage) {
			bar.setPercentage(percentage);
		}

		public void setBarVisible(boolean visible) {
			bar.setVisible(visible);
		}
		
		public void setIsHeading(boolean isHeading) {
			name.setFont(isHeading ? HEADING_FONT : NORMAL_FONT);
			value.setFont(isHeading ? HEADING_FONT : NORMAL_FONT);
		}

		@Override
		protected void paintComponent(Graphics g) {
			Graphics2D g2d = (Graphics2D) g;
			g2d.setColor(getBackground());
			g2d.fill(g.getClip());
		}
	}

	/**
	 * {@link JList} renderer that displays {@link PropertyItem} objects using
	 * {@link PropertyItemComponent}.
	 * 
	 * @author Peter Hofer
	 */
	protected static class PropertyItemRenderer implements ListCellRenderer {
		private PropertyItemComponent component = new PropertyItemComponent();

		private final Color ODD_ROWS_COLOR = Color.WHITE;
		private final Color EVEN_ROWS_COLOR = new Color(234, 234, 234); // bright gray

		@Override
		public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected,
				boolean cellHasFocus) {

			PropertyItem item = (PropertyItem) value;
			if (item != null) {
				component.setNameString(item.getName());
				component.setValueString(item.getValue());
				component.setIsHeading(item.isHeading());

				if (item.hasPercentage()) {
					component.setBarVisible(true);
					component.setBarPercentage(item.getPercentage());
					component.setBarColor(item.getColor());
				} else {
					component.setBarVisible(false);
				}

				if (index % 2 == 0) {
					component.setBackground(EVEN_ROWS_COLOR);
				} else {
					component.setBackground(ODD_ROWS_COLOR);
				}

				return component;
			}

			return null;
		}
	}

	private JList list;
	private PropertyItemRenderer renderer;

	/**
	 * Constructor.
	 */
	public PropertyListView() {
		setLayout(new BorderLayout());

		list = new JList();
		list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		list.setVisibleRowCount(-1);

		renderer = new PropertyItemRenderer();
		list.setCellRenderer(renderer);

		add(list, BorderLayout.CENTER);
	}

	/**
	 * Sets a list model that supplies {@link PropertyItem} objects.
	 * 
	 * @param model
	 *            List model to use.
	 */
	public void setModel(ListModel model) {
		list.setModel(model);
	}

	/**
	 * Copies a text representation of the contents of this summary view to the
	 * system clipboard.
	 */
	public void copyToClipboard() {
		ListModel model = list.getModel();
		StringBuilder text = new StringBuilder();

		int size = model.getSize();
		for (int i = 0; i < size; ++i) {
			PropertyItem item = (PropertyItem) model.getElementAt(i);
			text.append(item.getName());
			text.append(' ');
			text.append(item.getValue());
			text.append('\n');
		}

		Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
		clipboard.setContents(new StringSelection(text.toString()), null);
	}
}
