package gcspy.vis;

import gcspy.interpreter.Space;
import gcspy.interpreter.Stream;

import java.awt.BorderLayout;

import javax.swing.AbstractListModel;
import javax.swing.JComponent;

/**
 * Component that displays a summary of a space's streams using
 * {@link PropertyListView}.
 * 
 * @author Peter Hofer
 */
public class SpaceSummaryView extends JComponent {

	/**
	 * List Model that has a backing {@link Space} and generates
	 * {@link PropertyItem} objects for each stream and additionally for each
	 * enumeration value.
	 * 
	 * @author Peter Hofer
	 */
	private static class SpaceSummaryListModel extends AbstractListModel {

		private Space space;

		public SpaceSummaryListModel(Space space) {
			this.space = space;
		}

		public void setSpace(Space space) {
			this.space = space;
		}

		@Override
		public Object getElementAt(int index) {
			int i = 0;
			
			int streams = space.getStreamNum();
			for (int j = 0; j < streams; ++j) {
				Stream stream = space.getStream(j);
				
				if (i == index) {
					return SpaceSummaryGenerator.getStreamSummaryItem(stream);
				}
				i++;
				
				// For enumerations, enum values follow as single entries
				if (stream.getPresentation() == Stream.PRESENTATION_ENUM) {
					String[] names = stream.getEnumNames();
					if (index < i + names.length) {
						return SpaceSummaryGenerator.getStreamEnumSummaryItem(stream, index - i);
					}
					i += names.length;
				}
			}
			
			return null;
		}

		@Override
		public int getSize() {
			int size = 0;

			int streams = space.getStreamNum();
			for (int i = 0; i < streams; ++i) {
				Stream stream = space.getStream(i);
				size++;
				
				if (stream.getPresentation() == Stream.PRESENTATION_ENUM) {
					// Additional entry for each enum value
					String[] names = stream.getEnumNames();
					size += names.length;
				}
			}
			
			return size;
		}
	}

	private PropertyListView view;
	private SpaceSummaryListModel model;

	/**
	 * Constructor.
	 * 
	 * @param space
	 *            The space to display.
	 */
	public SpaceSummaryView(Space space) {
		if (space == null)
			throw new NullPointerException("space");

		setLayout(new BorderLayout());

		model = new SpaceSummaryListModel(space);

		view = new PropertyListView();
		view.setModel(model);
		add(view, BorderLayout.CENTER);
	}

	/**
	 * Sets the space displayed by this summary view.
	 * 
	 * @param space
	 *            Space to display.
	 */
	public void setSpace(Space space) {
		model.setSpace(space);
	}

	/**
	 * Updates the view.
	 */
	public void update() {
		repaint();
	}
}
