package gcspy.vis;

import gcspy.vis.utils.IconLoader;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * Component that consists of two buttons, one for zooming in and one for
 * zooming out. Clicking the buttons results in a change of the zoom value
 * within a defined range.
 * 
 * @author Peter Hofer
 * 
 */
public class ZoomComponent extends JComponent {

	private ActionListener buttonListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			int change;
			if (e.getSource() == zoomInB) {
				change = zoomInChange;
			} else {
				change = zoomOutChange;
			}

			int newValue = value + change;
			if (value < minimum) {
				value = minimum;
			} else if (value > maximum) {
				value = maximum;
			}

			if (newValue != value) {
				value = newValue;

				zoomInB.setEnabled(value != maximum);
				zoomOutB.setEnabled(value != minimum);

				fireStateChanged();
			}
		}
	};

	private JButton zoomOutB;
	private JButton zoomInB;

	private int zoomInChange = +1;
	private int zoomOutChange = -1;

	private int value;
	private int minimum;
	private int maximum;

	private List<ChangeListener> changeListeners;

	/**
	 * Initializes this zoom component to a value of 100, with a valid range of
	 * <i>0 <= x <= 200</i>.
	 * 
	 * @see #ZoomComponent(int, int, int)
	 */
	public ZoomComponent() {
		this(0, 200, 100);
	}

	/**
	 * Initializes this zoom component to the specified range (<i>min <= x <=
	 * max</i>) and initial value.
	 * 
	 * @param min
	 *            Minimum value.
	 * @param max
	 *            Maximum value.
	 * @param value
	 *            Initial value.
	 * @throws IllegalArgumentException
	 *             when <i>min > max</i> or <i>value < min</i> or <i>value >
	 *             max</i>.
	 */
	public ZoomComponent(int min, int max, int value) {
		if (min > max)
			throw new IllegalArgumentException("min > max");

		if (value < min || value > max)
			throw new IllegalArgumentException("value < min or value > max");

		this.minimum = min;
		this.maximum = max;
		this.value = value;

		setLayout(new BorderLayout());

		zoomInB = new JButton(IconLoader.getIcon(IconLoader.MAGNIFIER_LARGER_ICON));
		zoomInB.addActionListener(buttonListener);
		add(zoomInB, BorderLayout.WEST);

		zoomOutB = new JButton(IconLoader.getIcon(IconLoader.MAGNIFIER_SMALLER_ICON));
		zoomOutB.addActionListener(buttonListener);
		add(zoomOutB, BorderLayout.EAST);
	}

	/**
	 * Registers a change listener to be notified when the zoom level (value)
	 * changes.
	 * 
	 * @param tileSizeListener
	 *            Listener for zoom events.
	 */
	public void addChangeListener(ChangeListener tileSizeListener) {
		if (changeListeners == null) {
			changeListeners = new ArrayList<ChangeListener>();
		}
		changeListeners.add(tileSizeListener);
	}

	/**
	 * Returns the current value (zoom level).
	 */
	public int getValue() {
		return value;
	}

	private void fireStateChanged() {
		if (changeListeners != null) {
			ChangeEvent e = new ChangeEvent(this);
			for (ChangeListener l : changeListeners) {
				l.stateChanged(e);
			}
		}
	}

}
