/**
 ** $Id: OKCancelDialog.java 28 2005-06-20 13:13:35Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.vis.utils;

import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;

import net.miginfocom.swing.MigLayout;

/**
 * Dialog window that provides common functionality for OK/Cancel dialogs such
 * as a result code and predefined behavior.
 * 
 * @author Tony Printezis
 * @author Peter Hofer
 */
public abstract class OKCancelDialog extends JDialog {

	public static final int NO_RESULT = -1;
	public static final int RESULT_OK = 0;
	public static final int RESULT_CANCEL = 1;

	private ActionListener buttonListener = new ActionListener() {
		public void actionPerformed(ActionEvent event) {
			Object source = event.getSource();
			if (source == okB) {
				onOkClicked();
			} else if (source == cancelB) {
				onCancelClicked();
			}
		}
	};

	private JButton okB;
	private JButton cancelB;
	private JPanel buttonPanel;

	protected int dialogResult = NO_RESULT;

	/**
	 * Constructor.
	 * 
	 * @param owner
	 *            The parent window.
	 */
	public OKCancelDialog(Window owner) {
		super(owner);

		okB = Factory.createButton("OK", buttonListener);
		cancelB = Factory.createButton("Cancel", buttonListener);
	}

	/**
	 * Returns the dialog result, which can be {@link #RESULT_OK},
	 * {@link #RESULT_CANCEL}, {@link #NO_RESULT} or any user-defined integer.
	 */
	public int getDialogResult() {
		return dialogResult;
	}

	/**
	 * Provides subclasses with a panel that contains the Ok and Cancel buttons.
	 * Subsequent calls of this method will always return the exact same object.
	 * Subclasses may modify the panel at their own will by adding components or
	 * changing layout constraints.
	 */
	protected JPanel getButtonPanel() {
		if (buttonPanel == null) {
			buttonPanel = new JPanel(new MigLayout("fillx, insets 0", "[]", "[nogrid]"));
			buttonPanel.add(okB, "tag ok, sg button");
			buttonPanel.add(cancelB, "tag cancel, sg button");
		}
		return buttonPanel;
	}

	/**
	 * Called when the Cancel button has been clicked. The default
	 * implementation calls {@link #setDialogResult(int)} with
	 * {@link #RESULT_CANCEL} and then hides this dialog. Can be re-implemented
	 * in subclasses to alter the behavior.
	 */
	protected void onCancelClicked() {
		setDialogResult(RESULT_CANCEL);
		setVisible(false);
	}

	/**
	 * Called when the Ok button has been clicked. The default implementation
	 * calls {@link #setDialogResult(int)} with {@link #RESULT_OK} and then
	 * hides this dialog. Can be re-implemented in subclasses to alter the
	 * behavior.
	 */
	protected void onOkClicked() {
		setDialogResult(RESULT_OK);
		setVisible(false);
	}

	/**
	 * Sets the dialog result code to the specified value which can be one of
	 * the constants this class provides (e.g. {@link #RESULT_CANCEL}) or a
	 * custom result code.
	 */
	protected void setDialogResult(int result) {
		this.dialogResult = result;
	}
}
